@@ -1,4 +1,4 @@
-# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
+# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf-8 -*-
 #
 # Copyright 2002 Ben Escoto <ben@emerose.org>
 # Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
@@ -36,19 +36,16 @@ import sys
 import time
 import re
 import getpass
-import gettext
 import re
-import types
 import urllib.request  # pylint: disable=import-error
 import urllib.parse  # pylint: disable=import-error
 import urllib.error  # pylint: disable=import-error
 
 from duplicity import dup_temp
 from duplicity import file_naming
-from duplicity import globals
+from duplicity import config
 from duplicity import log
 from duplicity import path
-from duplicity import progress
 from duplicity import util
 
 from duplicity.util import exception_traceback
@@ -64,6 +61,7 @@ import duplicity.backends
 
 _backends = {}
 _backend_prefixes = {}
+_last_exception = None
 
 # These URL schemes have a backend with a notion of an RFC "network location".
 # The 'file' and 's3+http' schemes should not be in this list.
@@ -220,7 +218,7 @@ def get_backend(url_string):
 
     Raise InvalidBackendURL if the URL is not a valid URL.
     """
-    if globals.use_gio:
+    if config.use_gio:
         url_string = u'gio+' + url_string
     obj = get_backend_object(url_string)
     if obj:
@@ -296,6 +294,16 @@ class ParsedUrl(object):
         except Exception:
             raise InvalidBackendURL(u"Syntax error (hostname) in: %s" % url_string)
 
+        try:
+            self.query = pu.query
+        except Exception:
+            raise InvalidBackendURL(u"Syntax error (query) in: %s" % url_string)
+        if self.query:
+            self.query_args = urllib.parse.parse_qs(self.query)
+        else:
+            self.query = None
+            self.query_args = {}
+
         # init to None, overwrite with actual value on success
         self.port = None
         try:
@@ -305,9 +313,9 @@ class ParsedUrl(object):
             if self.scheme in [u'rclone']:
                 pass
             # old style rsync://host::[/]dest, are still valid, though they contain no port
-            elif not (self.scheme in [u'rsync'] and re.search(u'::[^:]*$', self.url_string)):
+            elif not (u'rsync' in self.scheme and re.search(u'::[^:]*$', self.url_string)):
                 raise InvalidBackendURL(u"Syntax error (port) in: %s A%s B%s C%s" %
-                                        (url_string, (self.scheme in [u'rsync']),
+                                        (url_string, (u'rsync' in self.scheme),
                                          re.search(u'::[^:]+$', self.netloc), self.netloc))
 
         # Our URL system uses two slashes more than urlparse's does when using
@@ -340,6 +348,9 @@ class ParsedUrl(object):
     def geturl(self):
         return self.url_string
 
+    def strip_auth(self):
+        return duplicity.backend.strip_auth_from_url(self)
+
 
 def strip_auth_from_url(parsed_url):
     u"""Return a URL from a urlparse object without a username or password."""
@@ -369,42 +380,55 @@ def retry(operation, fatal=True):
     # have to return a decorator function (which itself returns a function!)
     def outer_retry(fn):
         def inner_retry(self, *args):
-            for n in range(1, globals.num_retries + 1):
+            global _last_exception
+            errors_fatal, errors_default = config.are_errors_fatal.get(operation, (True, None))
+            for n in range(1, config.num_retries + 1):
                 try:
                     return fn(self, *args)
                 except FatalBackendException as e:
-                    # die on fatal errors
-                    raise e
-                except Exception as e:
-                    # retry on anything else
-                    log.Debug(_(u"Backtrace of previous error: %s")
-                              % exception_traceback())
-                    at_end = n == globals.num_retries
-                    code = _get_code_from_exception(self.backend, operation, e)
-                    if code == log.ErrorCode.backend_not_found:
-                        # If we tried to do something, but the file just isn't there,
-                        # no need to retry.
-                        at_end = True
-                    if at_end and fatal:
-                        def make_filename(f):
-                            if isinstance(f, path.ROPath):
-                                return util.escape(f.uc_name)
-                            else:
-                                return util.escape(f)
-                        extra = u' '.join([operation] + [make_filename(x) for x in args if (x and isinstance(x, str))])
-                        log.FatalError(_(u"Giving up after %s attempts. %s: %s")
-                                       % (n, e.__class__.__name__,
-                                          util.uexc(e)), code=code, extra=extra)
+                    _last_exception = e
+                    if not errors_fatal:
+                        # backend wants to report and ignore errors
+                        return errors_default
                     else:
-                        log.Warn(_(u"Attempt %s failed. %s: %s")
-                                 % (n, e.__class__.__name__, util.uexc(e)))
-                    if not at_end:
-                        if isinstance(e, TemporaryLoadException):
-                            time.sleep(3 * globals.backend_retry_delay)  # wait longer before trying again
+                        # die on fatal errors
+                        raise e
+                except Exception as e:
+                    _last_exception = e
+                    if not errors_fatal:
+                        # backend wants to report and ignore errors
+                        return errors_default
+                    else:
+                        # retry on anything else
+                        log.Debug(_(u"Backtrace of previous error: %s")
+                                  % exception_traceback())
+                        at_end = n == config.num_retries
+                        code = _get_code_from_exception(self.backend, operation, e)
+                        if code == log.ErrorCode.backend_not_found:
+                            # If we tried to do something, but the file just isn't there,
+                            # no need to retry.
+                            at_end = True
+                        if at_end and fatal:
+                            def make_filename(f):
+                                if isinstance(f, path.ROPath):
+                                    return util.escape(f.uc_name)
+                                else:
+                                    return util.escape(f)
+                            extra = u' '.join([operation] + [make_filename(x) for x in args
+                                                             if (x and isinstance(x, str))])
+                            log.FatalError(_(u"Giving up after %s attempts. %s: %s")
+                                           % (n, e.__class__.__name__,
+                                              util.uexc(e)), code=code, extra=extra)
                         else:
-                            time.sleep(globals.backend_retry_delay)  # wait a bit before trying again
-                        if hasattr(self.backend, u'_retry_cleanup'):
-                            self.backend._retry_cleanup()
+                            log.Warn(_(u"Attempt of %s Nr. %s failed. %s: %s")
+                                     % (fn.__name__, n, e.__class__.__name__, util.uexc(e)))
+                        if not at_end:
+                            if isinstance(e, TemporaryLoadException):
+                                time.sleep(3 * config.backend_retry_delay)  # wait longer before trying again
+                            else:
+                                time.sleep(config.backend_retry_delay)  # wait a bit before trying again
+                            if hasattr(self.backend, u'_retry_cleanup'):
+                                self.backend._retry_cleanup()
 
         return inner_retry
     return outer_retry
@@ -584,6 +608,22 @@ class BackendWrapper(object):
         else:
             raise NotImplementedError()
 
+    def pre_process_download(self, remote_filename):
+        u"""
+        Manages remote access before downloading files
+        (unseal data in cold storage for instance)
+        """
+        if hasattr(self.backend, u'pre_process_download'):
+            return self.backend.pre_process_download(remote_filename)
+
+    def pre_process_download_batch(self, remote_filenames):
+        u"""
+        Manages remote access before downloading files
+        (unseal data in cold storage for instance)
+        """
+        if hasattr(self.backend, u'pre_process_download_batch'):
+            return self.backend.pre_process_download_batch(remote_filenames)
+
     def delete(self, filename_list):
         u"""
         Delete each filename in filename_list, in order if possible.
