@@ -1,4 +1,4 @@
-# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
+# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf-8 -*-
 #
 # Copyright 2002 Ben Escoto <ben@emerose.org>
 # Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
@@ -20,26 +20,25 @@
 # Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 
 from builtins import map
-from builtins import filter
 from builtins import next
-from builtins import range
 from builtins import object
+from builtins import range
 
-import re  # @UnusedImport
-import types
-import os
+import re
 import sys
 import tempfile
 
-from duplicity import tarfile  # @UnusedImport
-from duplicity import librsync  # @UnusedImport
-from duplicity import log  # @UnusedImport
+from duplicity import errors
 from duplicity import diffdir
+from duplicity import config
+from duplicity import librsync
+from duplicity import log
 from duplicity import selection
+from duplicity import tarfile
 from duplicity import tempdir
-from duplicity import util  # @UnusedImport
-from duplicity.path import *  # @UnusedWildImport
-from duplicity.lazy import *  # @UnusedWildImport
+from duplicity import util
+from duplicity.lazy import *  # pylint: disable=unused-wildcard-import,redefined-builtin
+from duplicity.path import *  # pylint: disable=unused-wildcard-import,redefined-builtin
 
 u"""Functions for patching of directories"""
 
@@ -198,7 +197,7 @@ def get_index_from_tarinfo(tarinfo):
         if b'..' in index:
             raise PatchDirException(u"Tar entry %s contains '..'.  Security "
                                     u"violation" % util.fsdecode(tiname))
-    return (index, difftype, multivol)
+    return index, difftype, multivol
 
 
 class Multivol_Filelike(object):
@@ -214,7 +213,7 @@ class Multivol_Filelike(object):
         self.tarinfo_list = tarinfo_list  # must store as list for write access
         self.index = index
         self.buffer = b""
-        self.at_end = 0
+        self.at_end = False
 
     def read(self, length=-1):
         u"""Read length bytes from file"""
@@ -235,14 +234,13 @@ class Multivol_Filelike(object):
     def addtobuffer(self):
         u"""Add next chunk to buffer"""
         if self.at_end:
-            return None
-        index, difftype, multivol = get_index_from_tarinfo(  # @UnusedVariable
-            self.tarinfo_list[0])
+            return False
+        index, difftype, multivol = get_index_from_tarinfo(self.tarinfo_list[0])
         if not multivol or index != self.index:
             # we've moved on
             # the following communicates next tarinfo to difftar2path_iter
-            self.at_end = 1
-            return None
+            self.at_end = True
+            return False
 
         fp = self.tf.extractfile(self.tarinfo_list[0])
         self.buffer += fp.read()
@@ -252,24 +250,25 @@ class Multivol_Filelike(object):
             self.tarinfo_list[0] = next(self.tar_iter)
         except StopIteration:
             self.tarinfo_list[0] = None
-            self.at_end = 1
-            return None
-        return 1
+            self.at_end = True
+            return False
+        return True
 
     def close(self):
         u"""If not at end, read remaining data"""
         if not self.at_end:
-            while 1:
+            while True:
                 self.buffer = b""
                 if not self.addtobuffer():
                     break
-        self.at_end = 1
+        self.at_end = True
 
 
 class PathPatcher(ITRBranch):
     u"""Used by DirPatch, process the given basis and diff"""
     def __init__(self, base_path):
         u"""Set base_path, Path of root of tree"""
+        self.dir_basis_path = None
         self.base_path = base_path
         self.dir_diff_ropath = None
 
@@ -295,7 +294,7 @@ class PathPatcher(ITRBranch):
         if self.dir_diff_ropath:
             self.dir_diff_ropath.copy_attribs(self.dir_basis_path)
 
-    def can_fast_process(self, index, basis_path, diff_ropath):
+    def can_fast_process(self, index, basis_path, diff_ropath):  # pylint: disable=unused-argument
         u"""No need to recurse if diff_ropath isn't a directory"""
         return not (diff_ropath and diff_ropath.isdir())
 
@@ -345,7 +344,16 @@ class TarFile_FromFileobjs(object):
         u"""Set tarfile from next file object, or raise StopIteration"""
         if self.current_fp:
             assert not self.current_fp.close()
-        self.current_fp = next(self.fileobj_iter)
+
+        while True:
+            x = next(self.fileobj_iter)
+            if isinstance(x, errors.BadVolumeException):
+                # continue with the next volume
+                continue
+            else:
+                self.current_fp = x
+                break
+
         self.tarfile = util.make_tarfile(u"r", self.current_fp)
         self.tar_iter = iter(self.tarfile)
 
@@ -463,7 +471,7 @@ class IndexedTuple(object):
         elif isinstance(other, tuple):
             return self.data == other
         else:
-            return None
+            return False
 
     def __str__(self):
         return u"(%s).%s" % (u", ".join(map(str, self.data)), self.index)
@@ -608,7 +616,7 @@ class ROPath_IterWriter(ITRBranch):
                 ropath.copy(new_path)
 
         self.dir_new_path = self.base_path.new_index(index)
-        if self.dir_new_path.exists() and not globals.force:
+        if self.dir_new_path.exists() and not config.force:
             # base may exist, but nothing else
             assert index == (), index
         else:
@@ -620,7 +628,7 @@ class ROPath_IterWriter(ITRBranch):
         if self.dir_diff_ropath:
             self.dir_diff_ropath.copy_attribs(self.dir_new_path)
 
-    def can_fast_process(self, index, ropath):
+    def can_fast_process(self, index, ropath):  # pylint: disable=unused-argument
         u"""Can fast process (no recursion) if ropath isn't a directory"""
         log.Info(_(u"Writing %s of type %s") %
                  (util.fsdecode(ropath.get_relative_path()), ropath.type),
