@@ -1,4 +1,4 @@
-# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
+# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf-8 -*-
 #
 # Copyright 2002 Ben Escoto <ben@emerose.org>
 # Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
@@ -73,20 +73,18 @@ class Select(object):
     2 - the test says the file (must be directory) should be scanned
 
     Also, a selection function f has a variable f.exclude which should
-    be true iff f could potentially exclude some file.  This is used
+    be true if f could potentially exclude some file. This is used
     to signal an error if the last function only includes, which would
     be redundant and presumably isn't what the user intends.
 
     """
-    # This re should not match normal filenames, but usually just globs
-    glob_re = re.compile(u"(.*[*?[]|ignorecase\\:)", re.I | re.S)
-
     def __init__(self, path):
         u"""Initializer, called with Path of root directory"""
         assert isinstance(path, Path), str(path)
         self.selection_functions = []
         self.rootpath = path
         self.prefix = self.rootpath.uc_name
+        self.files_from = None
 
     def __iter__(self):  # pylint: disable=non-iterator-returned
         return self
@@ -128,18 +126,32 @@ class Select(object):
                          log.WarningCode.cannot_stat, util.escape(fullpath))
             return None
 
-        def diryield(path):
-            u"""Generate relevant files in directory path
+        def dir_scanner(path):
+            u"""Generator of files to be included under from filesystems
 
-            Returns (path, num) where num == 0 means path should be
-            generated normally, num == 1 means the path is a directory
-            and should be included iff something inside is included.
+            Yields (subpath, scan) where scan indicates that path is a
+            directory and should be included *if* something inside is
+            subsequently included (i.e. generated later)
+
+            This generator applies the configured selection functions and
+            so will not yield paths which are to be excluded.
+
+            If --files-from has been specified, this will yield paths from
+            that filelist only, subject to any selection functions.
 
             """
             # Only called by Iterate. Internal.
             # todo: get around circular dependency issue by importing here
             from duplicity import robust
-            for filename in robust.listpath(path):
+
+            if self.files_from is None:
+                files = robust.listpath(path)
+            elif path.uc_name in self.files_from:
+                files = self.files_from[path.uc_name]
+            else:
+                return
+
+            for filename in files:
                 new_path = robust.check_common_error(
                     error_handler, Path.append, (path, filename))
                 if new_path:
@@ -157,10 +169,10 @@ class Select(object):
                             diffdir.stats.Errors += 1
                     elif s == 1:
                         # Should be included
-                        yield (new_path, 0)
+                        yield (new_path, False)
                     elif s == 2 and new_path.isdir():
                         # Is a directory that should be scanned
-                        yield (new_path, 1)
+                        yield (new_path, True)
 
         if not path.type:
             # base doesn't exist
@@ -171,30 +183,30 @@ class Select(object):
         yield path
         if not path.isdir():
             return
-        diryield_stack = [diryield(path)]
-        delayed_path_stack = []
 
-        while diryield_stack:
+        dirs_to_scan = [dir_scanner(path)]
+        dirs_deferred = []
+        while dirs_to_scan:
             try:
-                subpath, val = next(diryield_stack[-1])
+                subpath, scan = next(dirs_to_scan[-1])
             except StopIteration:
-                diryield_stack.pop()
-                if delayed_path_stack:
-                    delayed_path_stack.pop()
+                dirs_to_scan.pop()
+                if dirs_deferred:
+                    dirs_deferred.pop()
                 continue
-            if val == 0:
-                if delayed_path_stack:
-                    for delayed_path in delayed_path_stack:
+            if scan:
+                dirs_deferred.append(subpath)
+                dirs_to_scan.append(dir_scanner(subpath))
+            else:
+                if dirs_deferred:
+                    for delayed_path in dirs_deferred:
                         log.Log(_(u"Selecting %s") % delayed_path.uc_name, 6)
                         yield delayed_path
-                    del delayed_path_stack[:]
+                    del dirs_deferred[:]
                 log.Debug(_(u"Selecting %s") % subpath.uc_name)
                 yield subpath
                 if subpath.isdir():
-                    diryield_stack.append(diryield(subpath))
-            elif val == 1:
-                delayed_path_stack.append(subpath)
-                diryield_stack.append(diryield(subpath))
+                    dirs_to_scan.append(dir_scanner(subpath))
 
     def Select(self, path):
         u"""Run through the selection functions and return dominant val 0/1/2"""
@@ -242,38 +254,80 @@ class Select(object):
         (option-string, (additional argument, filelist_fp)).
 
         """
+        # Sanity checks on --filter-* options for the benefit of users
+        if argtuples and argtuples[-1][0].startswith(u"--filter-"):
+            log.FatalError(_(u"""\
+The last file selection option is the filter option %s, which will have no
+effect as there are no subsequent file selection options. Exiting because
+this probably isn't what you meant.""") %
+                           (argtuples[-1][0],),
+                           log.ErrorCode.trailing_filter)
+        f_opt = set(opt[0] for opt in argtuples if opt[0].startswith(u"--filter-"))
+        f_def = (u"--filter-globbing", u"--filter-strictcase")
+        if f_opt and all(opt in f_def for opt in f_opt):
+            log.FatalError(_(u"""\
+Only these filter mode options were specified:
+    %s
+Case sensitive globbing is the default behaviour and so this has no effect.
+Exiting because this probably isn't what you meant.""") %
+                           (u", ".join(f_opt),),
+                           log.ErrorCode.redundant_filter)
+
         # Called by commandline.py set_selection. External.
         filelists_index = 0
+        mode = u"globbing"
+        no_case = False
         try:
             for opt, arg in argtuples:
                 if opt == u"--exclude":
-                    self.add_selection_func(self.glob_get_sf(arg, 0))
+                    self.add_selection_func(self.general_get_sf(arg, 0, mode, ignore_case=no_case))
                 elif opt == u"--exclude-if-present":
                     self.add_selection_func(self.present_get_sf(arg, 0), add_to_start=True)
                 elif opt == u"--exclude-device-files":
                     self.add_selection_func(self.devfiles_get_sf(), add_to_start=True)
-                elif (opt == u"--exclude-filelist") or (opt == u"--exclude-globbing-filelist"):
-                    # --exclude-globbing-filelist is now deprecated, as all filelists are globbing
-                    # but keep this here for the short term for backwards-compatibility
-                    for sf in self.filelist_globbing_get_sfs(filelists[filelists_index], 0, arg):
+                elif opt == u"--exclude-filelist":
+                    for sf in self.filelist_general_get_sfs(filelists[filelists_index], 0, arg, mode, no_case):
+                        self.add_selection_func(sf)
+                    filelists_index += 1
+                elif opt == u"--exclude-globbing-filelist":
+                    # --exclude-globbing-filelist is now deprecated, but if it
+                    # turns up then it needs to always work in globbing mode...
+                    for sf in self.filelist_general_get_sfs(filelists[filelists_index], 0, arg, u"globbing", no_case):
                         self.add_selection_func(sf)
                     filelists_index += 1
                 elif opt == u"--exclude-other-filesystems":
                     self.add_selection_func(self.other_filesystems_get_sf(0))
                 elif opt == u"--exclude-regexp":
-                    self.add_selection_func(self.regexp_get_sf(arg, 0))
+                    self.add_selection_func(self.regexp_get_sf(arg, 0, no_case))
                 elif opt == u"--exclude-older-than":
                     self.add_selection_func(self.exclude_older_get_sf(arg))
+                elif opt == u"--filter-strictcase":
+                    no_case = False
+                elif opt == u"--filter-globbing":
+                    mode = u"globbing"
+                elif opt == u"--filter-ignorecase":
+                    no_case = True
+                elif opt == u"--filter-literal":
+                    mode = u"literal"
+                elif opt == u"--filter-regexp":
+                    mode = u"regex"
+                elif opt == u"--files-from":
+                    self.parse_files_from(filelists[filelists_index], arg)
+                    filelists_index += 1
                 elif opt == u"--include":
-                    self.add_selection_func(self.glob_get_sf(arg, 1))
-                elif (opt == u"--include-filelist") or (opt == u"--include-globbing-filelist"):
-                    # --include-globbing-filelist is now deprecated, as all filelists are globbing
-                    # but keep this here for the short term for backwards-compatibility
-                    for sf in self.filelist_globbing_get_sfs(filelists[filelists_index], 1, arg):
+                    self.add_selection_func(self.general_get_sf(arg, 1, mode, no_case))
+                elif opt == u"--include-filelist":
+                    for sf in self.filelist_general_get_sfs(filelists[filelists_index], 1, arg, mode, no_case):
+                        self.add_selection_func(sf)
+                    filelists_index += 1
+                elif opt == u"--include-globbing-filelist":
+                    # --include-globbing-filelist is now deprecated, but if it
+                    # turns up then it needs to always work in globbing mode...
+                    for sf in self.filelist_general_get_sfs(filelists[filelists_index], 1, arg, u"globbing", no_case):
                         self.add_selection_func(sf)
                     filelists_index += 1
                 elif opt == u"--include-regexp":
-                    self.add_selection_func(self.regexp_get_sf(arg, 1))
+                    self.add_selection_func(self.regexp_get_sf(arg, 1, no_case))
                 else:
                     assert 0, u"Bad selection option %s" % opt
         except GlobbingError as e:
@@ -299,10 +353,56 @@ pattern (such as '**') which matches the base directory.""") %
         else:
             raise  # pylint: disable=misplaced-bare-raise
 
+    def parse_files_from(self, filelist_fp, list_name):
+        u"""Loads an explicit list of files to backup from a filelist, building
+        a dictionary of directories and their contents which can be used later
+        to emulate a filesystem walk over the listed files only.
+
+        Each specified path is unwound to identify the parents folder(s) as these
+        are implicitly to be included.
+
+        Paths read are not to be stripped, checked for comments, etc. Every
+        character on each line is significant and treated as part of the path.
+        """
+        # Internal. Used by ParseArgs.
+        log.Notice(_(u"Reading files to backup from list %s") % list_name)
+        separator = config.null_separator and u"\0" or u"\n"
+        filelist = {}
+        absolute_path = None
+        for line in filelist_fp.read().split(separator):
+            if not line:                # skip blanks
+                continue
+            if line.startswith(u"/"):   # no absolute paths thanks
+                absolute_path = line
+                break
+            while line:
+                dirname, basename = os.path.split(line)
+                path = os.path.join(self.rootpath.uc_name, dirname).rstrip(os.path.sep)
+                if path not in filelist:
+                    filelist[path] = set()
+                filelist[path].add(basename)
+                line = dirname
+
+        if absolute_path:
+            log.FatalError(_(u"""\
+Files-from list contains the absolute path:
+    %s
+All paths specified in a files-from list must be given relative to the backup
+source path.""") %
+                           (absolute_path,),
+                           log.ErrorCode.absolute_files_from)
+
+        if not filelist:
+            log.FatalError(_(u"""\
+Files-from list specified which contains no files, the backup will be empty as
+a result. Exiting as this probably isn't what you meant,"""), log.ErrorCode.empty_files_from)
+
+        self.files_from = {d: sorted(f) for d, f in filelist.items()}
+
     def parse_last_excludes(self):
         u"""Exit with error if last selection function isn't an exclude"""
         # Internal. Used by ParseArgs.
-        if (self.selection_functions and not self.selection_functions[-1].exclude):
+        if self.selection_functions and not self.selection_functions[-1].exclude:
             log.FatalError(_(u"""\
 Last selection expression:
     %s
@@ -322,12 +422,15 @@ probably isn't what you meant.""") %
 
     def filelist_sanitise_line(self, line, include_default):
         u"""
-        Sanitises lines of both normal and globbing filelists, returning (line, include) and line=None if blank/comment
+        Sanitises lines of both normal and globbing filelists, returning
+        (line, include) and line=None if blank/comment
 
-        The aim is to parse filelists in a consistent way, prior to the interpretation of globbing statements.
-        The function removes whitespace, comment lines and processes modifiers (leading +/-) and quotes.
+        The aim is to parse filelists in a consistent way, prior to the
+        interpretation of globbing statements. The function removes
+        whitespace, comment lines and processes modifiers (leading +/-)
+        and quotes.
         """
-        # Internal. Used by filelist_globbing_get_sfs
+        # Internal. Used by filelist_general_get_sfs
 
         line = line.strip()
         if not line:  # skip blanks
@@ -349,25 +452,28 @@ probably isn't what you meant.""") %
 
         return line, include
 
-    def filelist_globbing_get_sfs(self, filelist_fp, inc_default, list_name):
+    def filelist_general_get_sfs(self, filelist_fp, inc_default, list_name, mode=u"globbing", ignore_case=False):
         u"""Return list of selection functions by reading fileobj
 
         filelist_fp should be an open file object
-        inc_default is true iff this is an include list
+        inc_default is true if this is an include list
         list_name is just the name of the list, used for logging
-        See the man page on --[include/exclude]-globbing-filelist
+        mode indicates whether to glob, regex, or not
 
         """
         # Internal. Used by ParseArgs.
-        log.Notice(_(u"Reading globbing filelist %s") % list_name)
+        log.Notice(_(u"Reading %s filelist %s") % (mode, list_name))
         separator = config.null_separator and u"\0" or u"\n"
-        filelist_fp.seek(0)
+        try:
+            filelist_fp.seek(0)
+        except:
+            pass
         for line in filelist_fp.read().split(separator):
             line, include = self.filelist_sanitise_line(line, inc_default)
             if not line:
                 # Skip blanks and comment lines
                 continue
-            yield self.glob_get_sf(line, include)
+            yield self.general_get_sf(line, include, mode, ignore_case)
 
     def other_filesystems_get_sf(self, include):
         u"""Return selection function matching files on other filesystems"""
@@ -385,12 +491,17 @@ probably isn't what you meant.""") %
         sel_func.name = u"Match other filesystems"
         return sel_func
 
-    def regexp_get_sf(self, regexp_string, include):
+    def regexp_get_sf(self, regexp_string, include, ignore_case=False):
         u"""Return selection function given by regexp_string"""
         # Internal. Used by ParseArgs and unit tests.
         assert include == 0 or include == 1
+
+        flags = 0
+        if ignore_case:
+            flags = re.IGNORECASE
+
         try:
-            regexp = re.compile(regexp_string)
+            regexp = re.compile(regexp_string, flags)
         except Exception:
             log.Warn(_(u"Error compiling regular expression %s") % regexp_string)
             raise
@@ -402,7 +513,8 @@ probably isn't what you meant.""") %
                 return None
 
         sel_func.exclude = not include
-        sel_func.name = u"Regular expression: %s" % regexp_string
+        sel_func.name = u"regular expression %s %scase: %s" % \
+            (include and u"include" or u"exclude", ignore_case and u"no-" or u"", regexp_string)
         return sel_func
 
     def devfiles_get_sf(self):
@@ -418,31 +530,47 @@ probably isn't what you meant.""") %
         sel_func.name = u"Exclude device files"
         return sel_func
 
-    def glob_get_sf(self, glob_str, include):
-        u"""Return selection function given by glob string"""
-        # Internal. Used by ParseArgs, filelist_globbing_get_sfs and unit tests.
-        assert include == 0 or include == 1
-        assert isinstance(glob_str, str)
-        if glob_str == u"**":
-            sel_func = lambda path: include
-        else:
-            sel_func = self.glob_get_normal_sf(glob_str, include)
+    def general_get_sf(self, pattern_str, include, mode=u"globbing", ignore_case=False):
+        u"""Return selection function given by a pattern string
 
-        sel_func.exclude = not include
-        sel_func.name = u"Command-line %s glob: %s" % \
-                        (include and u"include" or u"exclude", glob_str)
-        return sel_func
+        The selection patterns are interpretted in accordance with the mode
+        argument, "globbing", "literal", or "regex".
+
+        The 'ignorecase:' prefix is a legacy feature which historically lived on
+        the globbing code path and was only ever documented as working for globs.
+        """
+
+        # Internal. Used by ParseArgs, filelist_general_get_sfs and unit tests.
+        assert include == 0 or include == 1
+        assert isinstance(pattern_str, str)
+
+        # legacy prefix applies *only* in globbing mode
+        if mode == u"globbing" and pattern_str.lower().startswith(u"ignorecase:"):
+            pattern_str = pattern_str[len(u"ignorecase:"):]
+            pattern_str = util.casefold_compat(util.fsdecode(pattern_str))
+            ignore_case = True
+
+        if mode == u"globbing":
+            return self.glob_get_sf(pattern_str, include, ignore_case)
+        elif mode == u"literal":
+            return self.literal_get_sf(pattern_str, include, ignore_case)
+        elif mode == u"regex":
+            return self.regexp_get_sf(pattern_str, include, ignore_case)
+        else:
+            assert 0, u"Bad selection mode %s" % mode
 
     def present_get_sf(self, filename, include):
         u"""Return selection function given by existence of a file in a directory"""
         # Internal. Used by ParseArgs.
         assert include == 0 or include == 1
 
+        # todo: get around circular dependency issue by importing here
+        from duplicity.robust import check_common_error
+
         def exclude_sel_func(path):
             # do not follow symbolic links when checking for file existence!
             if path.isdir():
-                # First check path is read accessible
-                if not (os.access(path.name, os.R_OK)):
+                def error_handler(_exc, _filename):
                     # Path is not read accessible
                     # ToDo: Ideally this error would only show if the folder
                     # was ultimately included by the full set of selection
@@ -453,7 +581,8 @@ probably isn't what you meant.""") %
                         log.WarningCode.cannot_read, util.escape(path.uc_name))
                     if diffdir.stats:
                         diffdir.stats.Errors += 1
-                elif path.contains(filename):
+                    return False
+                if check_common_error(error_handler, path.contains, [filename]):
                     return 0
                 else:
                     return None
@@ -469,39 +598,42 @@ probably isn't what you meant.""") %
                         (include and u"include-if-present" or u"exclude-if-present", filename)
         return sel_func
 
-    def glob_get_normal_sf(self, glob_str, include):
-        u"""Return selection function based on glob_str
-
-        The basic idea is to turn glob_str into a regular expression,
-        and just use the normal regular expression.  There is a
-        complication because the selection function should return '2'
-        (scan) for directories which may contain a file which matches
-        the glob_str.  So we break up the glob string into parts, and
-        any file which matches an initial sequence of glob parts gets
-        scanned.
-
-        Thanks to Donovan Baarda who provided some code which did some
-        things similar to this.
-
-        """
+    def glob_get_sf(self, glob_str, include, ignore_case=False):
+        u"""Return selection function based on glob_str"""
         assert isinstance(glob_str, str), \
             u"The glob string " + glob_str.decode(sys.getfilesystemencoding(), u"ignore") + u" is not unicode"
-        ignore_case = False
-
-        if glob_str.lower().startswith(u"ignorecase:"):
-            # Glob string starts with ignorecase, so remove that from the
-            # string and change it to lowercase.
-            glob_str = glob_str[len(u"ignorecase:"):].lower()
-            ignore_case = True
 
         # Check to make sure prefix is ok, i.e. the glob string is within
         # the root folder being backed up
-        file_prefix_selection = select_fn_from_glob(glob_str, include=1)(self.rootpath)
-        if not file_prefix_selection:
+        if not select_fn_from_glob(glob_str, 1, ignore_case)(self.rootpath):
             # file_prefix_selection == 1 (include) or 2 (scan)
             raise FilePrefixError(glob_str)
 
-        return select_fn_from_glob(glob_str, include, ignore_case)
+        if glob_str == u"**":
+            sel_func = lambda path: include
+        else:
+            sel_func = select_fn_from_glob(glob_str, include, ignore_case)
+
+        sel_func.exclude = not include
+        sel_func.name = u"shell glob %s %scase: %s" % \
+                        (include and u"include" or u"exclude", ignore_case and u"no-" or u"", glob_str)
+
+        return sel_func
+
+    def literal_get_sf(self, lit_str, include, ignore_case=False):
+        u"""Return a selection function that matches a literal string while
+        still including the contents of any folders which are matched
+        """
+
+        # Check to make sure prefix is ok, see also glob_get_sf()
+        if not self.select_fn_from_literal(lit_str, 1, ignore_case)(self.rootpath):
+            raise FilePrefixError(lit_str)
+
+        sel_func = self.select_fn_from_literal(lit_str, include, ignore_case)
+        sel_func.exclude = not include
+        sel_func.name = u"literal string %s %scase: %s" % \
+                        (include and u"include" or u"exclude", ignore_case and u"no-" or u"", lit_str)
+        return sel_func
 
     def exclude_older_get_sf(self, date):
         u"""Return selection function based on files older than modification date """
@@ -520,3 +652,37 @@ probably isn't what you meant.""") %
         sel_func.exclude = True
         sel_func.name = u"Select older than %s" % (date,)
         return sel_func
+
+    def select_fn_from_literal(self, lit_str, include, ignore_case=False):
+        u"""Return a function test_fn(path) which test where a path matches a
+        literal string. See also select_fn_from_blog() in globmatch.py
+
+        This function is separated from literal_get_sf() so that it can be used
+        to test the prefix without creating a loop.
+
+        TODO: this doesn't need to be part of the Select class type, but not
+        sure where else to put it?
+        """
+        if lit_str != u"/" and lit_str[-1] == u"/":
+            lit_str = lit_str[:-1]
+
+        if ignore_case:
+            lit_str = util.casefold_compat(util.fsdecode(lit_str))
+
+        def test_fn(path):
+            # FIXME: caller to do this once, rather than for every path?
+            if ignore_case:
+                uc_name = util.casefold_compat(util.fsdecode(path.uc_name))
+            else:
+                uc_name = path.uc_name
+
+            if uc_name == lit_str:
+                return include
+            elif uc_name.startswith(lit_str) and uc_name[len(lit_str)] == u"/":
+                return include
+            elif lit_str.startswith(uc_name) and lit_str[len(uc_name)] == u"/" and include == 1:
+                return 2
+            else:
+                return None
+
+        return test_fn
